import getpass
import os
import random
import readline
import socket
import string
import sys
from datetime import datetime
from signal import SIG_DFL, SIGPIPE, signal
from typing import Any, Dict, Union

from novem.exceptions import Novem401

from ..api_ref import NovemAPI
from ..utils import cl, colors, get_current_config
from ..version import __version__
from .config import check_if_profile_exists, update_config
from .mail import mail
from .plot import plot
from .setup import setup

sys.tracebacklimit = 0


def input_with_prefill(prompt: str, text: str) -> Any:
    def hook() -> Any:
        readline.insert_text(text)
        readline.redisplay()

    readline.set_pre_input_hook(hook)
    result = input(prompt)
    readline.set_pre_input_hook()
    return result


def do_update_config(
    profile: str,
    username: str,
    api_root: str,
    token_name: str,
    token: str,
    path: str,
) -> None:
    (status, path) = update_config(
        profile, username, api_root, token_name, token, path
    )

    print(
        f"{cl.OKGREEN} \u2713 {cl.ENDC}new token "
        f'{cl.OKCYAN}"{token_name}"{cl.ENDC} created and'
        f" saved to {path}"
    )

    # save file


def init_config(args: Dict[str, Any] = None) -> None:
    """
    Initialize user and config

    the --init flag has been supplied, this means we are going to update the
    configuration file. There are several different scenarios we have to adapt
    to in this case:

    * a simple --init with no options and missing config file
      * request username and password from user
      * authenticate against service
      * create config file and store credentials if successful

    * a simple --init with no options and existing config file
      * check if config has a default user
      * if default user exist in config, inform user that config
      * already exist and terminate

    * a simple --init with --profile <username> and missing config file
      * request username and password from user
        * with username prefilled as --profile
      * authenticate against service
      * create config file and store credentials if successful

      --init

      --profile
      --token
      --api-url
      --force

      user in config

      config exist



      if --token not supplied, request username and password
      if --profile specified, prefill username

      if token supplied request token info, if username password

    """

    if not args:
        # make mypy happy, return if no argument supplied
        return

    token: Union[str, None] = None

    if "token" in args:
        token = args["token"]

    api_root: str = args["api-url"]
    profile: str = args["profile"]
    force: str = args["force"]
    config_path: str = args["config_path"]

    # first check if we have a valid config
    profile_exists: bool = check_if_profile_exists(profile, config_path)
    if profile_exists and not force:
        print(
            f"{cl.WARNING} ! {cl.ENDC}"
            f' The supplied profile "{cl.OKCYAN}{profile}{cl.ENDC}" already '
            f"exist, use --force to override"
        )
        sys.exit(1)

    valid_char_sm = string.ascii_lowercase + string.digits
    valid_char = valid_char_sm + "-_"
    hostname: str = socket.gethostname()

    token_name: Union[str, None] = None
    if not token_name:
        token_hostname: str = "".join(
            [x for x in hostname.lower() if x in valid_char]
        )
        nounce: str = "".join(random.choice(valid_char_sm) for _ in range(8))
        token_name = f"novem-python-{token_hostname}-{nounce}"

    new_token_name = "".join([x for x in token_name if x in valid_char])

    if token_name != new_token_name:
        print(
            f"{cl.WARNING} ! {cl.ENDC}"
            "The supplied token name contained invalid charracters,"
            f' token changed to "{cl.OKCYAN}{new_token_name}{cl.ENDC}"'
        )
        token_name = new_token_name

    # get novem username
    prefill = ""
    if "profile" in args:
        prefill = args["profile"]

    username = input_with_prefill(" \u2022 novem.no username: ", prefill)
    # username = "abc"

    # get novem password
    password = getpass.getpass(" \u2022 novem.no password: ")

    # authenticate and request token by name
    req = {
        "username": username,
        "password": password,
        "token_name": token_name,
        "token_description": (
            f'cli token created for "{hostname}" '
            f'on "{datetime.now():%Y-%m-%d:%H:%M:%S}"'
        ),
    }

    if not api_root:
        (hasconf, curconf) = get_current_config(**args)
        # if our config exist, try to read it from there
        if not hasconf:
            api_root = "https://api.novem.no/v1/"
        else:
            api_root = curconf["api_root"]

    # let's grab our token
    ignore_ssl = False
    if "ignore_ssl" in args:
        ignore_ssl = args["ignore_ssl"]

    novem = NovemAPI(
        api_root=api_root, ignore_config=True, ignore_ssl=ignore_ssl
    )

    try:
        res = novem.create_token(req)
        token = res["token"]
        token_name = res["token_name"]

    except Novem401:
        print("Invalid username and/or password")
        sys.exit(1)

    # default profile to username if not supplied
    if not profile:
        profile = username

    # let's write our config
    do_update_config(
        profile, username, api_root, token_name, token, config_path
    )


def print_short(parser: Any) -> None:
    parser.print_usage()

    print()
    print("Novem commandline interface - no options supplied")
    print()
    print("  novem -h, --help      print a concise help")
    print("  novem --guide         print the comprehensive novem cli guide")
    print()
    print(
        "  novem --init          if you already have an account and want"
        " to get started"
    )
    print()
    print("  novem -p              list your plots")


def run_cli_wrapped() -> None:
    colors()

    raw_args = sys.argv[1:]

    #    return

    # Gather the provided arguements as an array.
    # (parser:Any, args:Dict[str, str]) = setup(raw_args)
    (parser, args) = setup(raw_args)

    if len(raw_args) == 0:
        print_short(parser)
        return

    if args and args["version"]:
        print(f"novem {__version__}")
        return

    # we are getting an init instruction
    if args and args["init"]:
        init_config(args)
        return

    # verify profile
    if args and args["profile"]:
        config_path: str = args["config_path"]
        profile_exists: bool = check_if_profile_exists(
            args["profile"], config_path
        )
        if not profile_exists:
            print(
                f'Profile "{args["profile"]}" doens\'t exist in your config. '
                "Please add it using:"
            )
            print(f'novem --init --profile {args["profile"]}')

            sys.exit(1)

    # check info and if present get info
    if args and args["info"]:
        novem = NovemAPI(**args)
        info = novem.read("/whoami")
        print(info, end="")
        return

    # if --fs is set get terminal dimensions and ammend qpr
    if args and "fs" in args and args["fs"]:
        sz = os.get_terminal_size()
        qpr = ""
        if "qpr" in args and args["qpr"]:
            qpr = f"{args['qpr']},"

        qpr = f"{qpr}cols={sz.columns},rows={sz.lines-1}"
        args["qpr"] = qpr

    # operate on plot
    if args and args["plot"] != "":
        plot(args)
    elif args and args["mail"] != "":
        mail(args)
        pass


def run_cli() -> None:
    signal(SIGPIPE, SIG_DFL)  # supress broken pipe error
    try:
        run_cli_wrapped()
    except KeyboardInterrupt:
        pass


__all__ = ["run_cli"]



import os
import sys
from typing import Any, Dict

from novem.exceptions import Novem404

from .. import Mail
from ..api_ref import NovemAPI
from ..utils import data_on_stdin
from .editor import edit
from .vis import list_vis, list_vis_shares


def mail(args: Dict[str, Any]) -> None:

    # we are invoked so plot must exist
    mail_name = args["mail"]

    if mail_name is None:
        # we need to list plots
        list_vis(args, "Mail")
        return

    # if delete flag is set, we need to delete it
    if args["delete"] and not args["share"]:
        # creating a plot just to delete it seems wasteful
        # We'll just use the raw api
        novem = NovemAPI(**args)

        try:
            novem.delete(f"vis/mails/{mail_name}")
            return
        except Novem404:
            print(f"Mail {mail_name} did not exist")
            sys.exit(1)

    usr = None
    if "for_user" in args and args["for_user"]:
        usr = args["for_user"]

    # we are going to modify one of our plots, listing and deletion has
    # already been handled so we can use the plot api without wasting
    # "resources'

    # this will create the plot if it does not exist
    ignore_ssl = False
    if "ignore_ssl" in args:
        ignore_ssl = args["ignore_ssl"]

    create = args["create"]

    # let's add our to, cc, subject
    p = Mail(
        mail_name,
        user=usr,
        ignore_ssl=ignore_ssl,
        create=create,
        config_path=args["config_path"],
        to=args["to"],
        cc=args["cc"],
        bcc=args["bcc"],
        subject=args["subject"],
        qpr=args["qpr"],
        debug=args["debug"],
    )

    # this is a data dump instruction, we'll ignore everything else
    if "dump" in args and args["dump"]:
        path = args["dump"]

        print(f'Dumping api tree structure to "{path}"')
        p.api_dump(outpath=path)
        return

    # if we detect a tree query then we'll discard all other IO
    if "tree" in args and args["tree"] != -1:
        path = args["tree"]
        if not path:
            path = "/"

        ts = p.api_tree(colors=True, relpath=path)
        print(ts)
        return

    # if we have the -e or edit flag then this takes presedence over all other
    # inputs
    if "edit" in args and args["edit"]:
        path = args["edit"]

        # fetch our target and warn if it doens't exist
        ctnt = p.api_read(f"/{path}")

        # get new content
        nctnt = edit(contents=ctnt, use_tty=True)

        if ctnt != nctnt:
            # update content
            p.api_write(f"/{path}", nctnt)

    else:

        ptype = args["type"]
        # print(ptype)
        if ptype:
            p.type = ptype

        found_stdin = False
        stdin_has_data, stdin_data = data_on_stdin()

        # check if we have any explicit inputs [-w's]
        if args["input"] and len(args["input"]):
            # we have inputs
            for i in args["input"]:
                path = f"/{i[0]}"

                if len(i) == 1:
                    if stdin_has_data and not found_stdin:
                        ctnt = stdin_data

                        p.api_write(path, ctnt)
                        found_stdin = True
                    elif found_stdin:
                        print(
                            "stdin can only be sent to a single destination"
                            " per invocation"
                        )
                        sys.exit(1)
                    else:
                        print(
                            f'No data found on stdin, "-w {path}" requires'
                            f" data to be supplied on stdin"
                        )
                        sys.exit(1)

                elif len(i) == 2 and i[1][0] == "@":
                    try:
                        fn = os.path.expanduser(i[1][1:])
                        with open(fn, "r") as f:
                            ctnt = f.read()
                            p.api_write(path, ctnt)
                    except FileNotFoundError:
                        print(
                            f'The supplied input file "{fn}" does not exist.'
                            f" Please review your options"
                        )
                        sys.exit(1)

                else:
                    ctnt = i[1]
                    p.api_write(path, ctnt)

        # check if we have standard input and not already "used" it
        if not found_stdin and stdin_has_data:
            ctnt = stdin_data
            # got stdin data
            p.content = ctnt

    # check if we are changing any permissions
    share: str = args["share"]

    if share is not None and share != "" and args["create"]:
        # add a share to the plot

        p.shared += share  # type: ignore

    if share is not None and share != "" and args["delete"]:
        # remove a share from the mail

        p.shared -= share  # type: ignore

    # check if we should print our shares, we will not provide other outputs
    if share is None:
        list_vis_shares(mail_name, args, "Mail")

        return

    # sending, check our T and S
    if args["test"]:
        p.test()
    elif args["send"]:
        p._send()

    # Output - we only allow a singular -o or -x and will return as soon as
    # we find one

    # -x takes presedence
    x = args["tc"]
    if x:
        # get file endpoint
        print(p.files.ansi, end="")  # type: ignore

        # exit as we only support one output
        return

    # TODO: check if we are reading any valus from the mail [-r or -x]
    out = args["out"]
    if out:
        outp = p.api_read(f"/{out}")
        print(outp, end="")



from typing import Any, Optional

from novem.exceptions import NovemException


class NovemSelectorException(NovemException):
    def __init__(self, message: str):

        message = f"Invalid selector format {message}."

        super().__init__(message)


class Selector(object):
    """
    A novem selector is used for selecting a subset of a table plot
    and applying a style to it
    """

    def __init__(
        self,
        selector: Any,
        applicator: Any,
        r: Any = None,
        c: Optional[str] = None,
        i: Optional[str] = None,
        co: Optional[int] = None,
        io: Optional[int] = None,
    ) -> None:
        """ """

        # TODO: verify that it's a valid novem selector before
        # storing it and throw an error if not
        self.selector = selector
        self.applicator = applicator
        self.ref = r
        self.c = c
        self.i = i
        self.co = co
        self.io = io

    def get_selector_string(self) -> str:
        """ """
        if isinstance(self.selector, str):
            return self.selector

        cols: str = ""
        rows: str = ""

        # if we don't have a columns attribute assume we're a series
        columns: Any = getattr(self.selector, "columns", None)
        index: Any = getattr(self.selector, "index", None)

        if self.ref is None:
            # TODO: Raise exception
            raise NovemSelectorException(
                "We need a reference to correctly infer pandas selections"
            )

        if index is None:
            raise NovemSelectorException(
                "There is no data in our dataframe, we need at "
                "least one value."
            )

        cola = 1
        if self.co:
            cola = cola + self.co

        if columns is None:
            # grab column index of our series name
            try:
                cols = self.ref.columns.get_loc(self.selector.name) + cola
            except KeyError:
                cols = ":"
        else:

            if list(self.selector.columns.values) == list(
                self.ref.columns.values
            ):
                # cols = ":"
                cols = ",".join(
                    [
                        str(self.ref.columns.get_loc(x) + cola)
                        for x in self.selector.columns.values
                    ]
                )
            else:
                cols = ",".join(
                    [
                        str(self.ref.columns.get_loc(x) + cola)
                        for x in self.selector.columns.values
                    ]
                )

        ila = 1
        if self.io:
            ila = ila + self.io

        if list(self.selector.index.values) == list(self.ref.index.values):
            rows = ":"
        else:
            try:
                rows = ",".join(
                    [
                        str(self.ref.index.get_loc(x) + ila)
                        for x in self.selector.index.values
                    ]
                )
            except KeyError:
                rows = self.ref.index.get_loc(self.selector.name) + ila

        if self.c:
            cols = self.c

        if self.i:
            rows = self.i

        return f"{rows} {cols}"

    def __str__(self) -> str:
        """
        return a string representaion of our selector
        """

        return f"{self.get_selector_string()} {self.applicator}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['novem',
 'novem.cli',
 'novem.colors',
 'novem.exceptions',
 'novem.table',
 'novem.vis']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.26.0,<3.0.0']

entry_points = \
{'console_scripts': ['novem = novem.cli:run_cli']}

setup_kwargs = {
    'name': 'novem',
    'version': '0.3.0',
    'description': 'python library for the novem.no data visualisation platform',
    'long_description': '\n# novem - data visualisation for coders\n\nA wrapper library for the novem.no data visualisation platform. Create charts,\ndocuments e-mails and dashboards through one simple api.\n\n**NB:** novem is currently in closed alpha, if you want to try it out please\nreach out to hello@novem.no\n\n\n## Exampels\n\nCreate a linechart from a dataframe using pandas data reader\n\n```python\nfrom pandas_datareader import data\nfrom novem import Plot\n\nline = Plot("aapl_price_hist", type="line", name="Apple price history")\n\n# Only get the adjusted close.\naapl = data.DataReader("AAPL",\n                       start="2015-1-1",\n                       end="2021-12-31",\n                       data_source="yahoo")["Adj Close"]\n\n# send data to the plot\naapl.pipe(line)\n\n# url to view plot\nprint(line.url)\n```\n\n\n## Getting started\nTo get started with novem you\'ll have to register an account, currently this\ncan be done by reaching out to the novem developers on hello@novem.no.\n\nOnce you have a username and password you can setup your environment using:\n```bash\n  python -m novem --init\n```\n\nIn additon to invoking the novem module as shown above, the novem package also\nincludes an extensive command-line interface (cli). Check out CLI.md in this\nrepostiory or [novem.no](https://novem.no) for more details.\n\n\n\n## Creating a plot\nNovem represents plots as a Plot class that can be imported from the main novem\npackage `from novem import Plot`.\n\nThe plot class takes a single mandatory positional argument, the name of the\nplot.\n * If the plot name is new, the instantiation of the class will create the plot.\n * If the plot name already exist, then the new object will operate on the\n   existing plot.\n\nIn addition to the name, there are two broad categories of options for a\nplot, data and config:\n * The **data** contains the actual information to visualise (usually in the form\n   of numeric csv)\n * **Config**, which contains information about the visual such as:\n   * Type (bar, line, donut, map etc)\n   * Titles/captions/names/colors/legends/axis etc\n\n\nThere are two ways to interact with the plots, one can either supply all\nthe neccessary options as named arguments when creating the plot, or use the\nproperty accessors to modfity them one by one (this is more helpful when working\nwith the plot in an interactive way). Below is an example of the two\napproaches.\n\n```python\nfrom novem import Plot\n\n# everything in the constructor\nbarchart = Plot(<name>, \\\n  type="bar", \\\n  title="barchart title", \\\n  caption = "caption"\n)\n\n# property approach\nbarchart = Plot("plot_name")\nbarchart.type = "bar"\nbarchart.title = "barchart title"\nbarchart.caption = "caption"\n```\n\nIn addition to setting individual properties, the plot object is also callable.\nThis means that the resulting plot can be used as a function, either by being\nprovided data as an argument, or used as part of a pipe chain.\n\n```python\nfrom novem import Plot\nimport pandas as pd\nimport numpy as np\n\n# construct some random sample data\ndf = pd.DataFrame(np.random.randn(100, 4), columns=list("ABCD")).cumsum()\n\nline = Plot("new_line", type="line")\n\n# alternative one, setting data explicitly to a csv string\nline.data = df.to_csv()\n\n# or let the plot invoke the to_csv\nline.data = df\n\n# alternative two, calling Plot with a csv string\nline(df.to_csv())\n\n# alternative three calling the Plot with an object that has a to_csv function\nline(df)\n\n# or\ndf.pipe(line)\n\n```\n\n\n**NB:** All novem plot operations are live.\nThis means that as soon as you write to or modify any aspects of the plot\nobject, those changes are reflected on the novem server and anyone watching\nthe plot in real time.\n\n\n\n## Contribution and development\nThe novem python library and platform is under active development, contributions\nor issues are most welcome.\n\nFor guidelines on how to contribute, please check out the CONTRIBUTING.md file\nin this repository.\n\n\n## LICENSE\nThis python library is licensed under the MIT license, see the LICENSE file for\ndetails\n',
    'author': 'Sondov Engen',
    'author_email': 'sondov@novem.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://novem.no',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import configparser
import io
import os
import platform
import select
import sys
import unicodedata
from typing import Any, Dict, List, Tuple, Union

NOVEM_PATH = "novem"
NOVEM_NAME = "novem.conf"


class cl:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    FGGRAY = "\033[38;5;246m"
    BGGRAY = "\033[48;5;234m"


def disable_colors() -> None:
    cl.HEADER = ""
    cl.OKBLUE = ""
    cl.OKCYAN = ""
    cl.OKGREEN = ""
    cl.WARNING = ""
    cl.FAIL = ""
    cl.ENDC = ""
    cl.BOLD = ""
    cl.UNDERLINE = ""
    cl.FGGRAY = ""
    cl.BGGRAY = ""


def colors() -> None:
    # ignore color disable if --colors in argv
    for a in sys.argv:
        if a == "--color":
            return

    # disable colors if not supported
    for handle in [sys.stdout, sys.stderr]:
        if (hasattr(handle, "isatty") and handle.isatty()) or (
            "TERM" in os.environ and os.environ["TERM"] == "ANSI"
        ):
            if platform.system() == "Windows" and not (
                "TERM" in os.environ and os.environ["TERM"] == "ANSI"
            ):
                disable_colors()
        else:
            disable_colors()


def get_user_config_directory() -> Union[str, None]:
    """Returns a platform-specific root directory for user config settings."""
    # On Windows, prefer %LOCALAPPDATA%, then %APPDATA%, since we can expect
    # the AppData directories to be ACLed to be visible only to the user and
    # admin users (https://stackoverflow.com/a/7617601/1179226). If neither is
    # set, return None instead of falling back to something that may be
    # world-readable.
    if os.name == "nt":
        appdata = os.getenv("LOCALAPPDATA")
        if appdata:
            return appdata
        appdata = os.getenv("APPDATA")
        if appdata:
            return appdata
        return None
    # On non-windows, use XDG_CONFIG_HOME if set, else default to ~/.config.
    xdg_config_home = os.getenv("XDG_CONFIG_HOME")
    if xdg_config_home:
        return xdg_config_home
    return os.path.join(os.path.expanduser("~"), ".config")


def get_config_path() -> Tuple[str, str]:
    """
    Get default configuration path
    """

    config_path: Union[str, None] = get_user_config_directory()
    novem_dir = f"{config_path}/{NOVEM_PATH}"
    novem_config = f"{config_path}/{NOVEM_PATH}/{NOVEM_NAME}"

    return (novem_dir, novem_config)


def get_current_config(
    **kwargs: Any,
) -> Tuple[bool, Dict[str, Any]]:
    """
    Resolve and return the current config options

    Contains :
    current user
    current profile
    current token
    current api_root
    """

    # config path can be supplied as an option, if it is use that
    if "config_path" not in kwargs or not kwargs["config_path"]:
        (novem_dir, config_path) = get_config_path()
    else:
        config_path = kwargs["config_path"]

    co: Dict[str, Any] = {}

    # defaults
    co["ignore_ssl_warn"] = False

    # in addition, we can be instructed to ignore the config
    # if we are ignoring the config then the api root must be provided
    if "ignore_config" in kwargs:
        co["api_root"] = kwargs["api_root"]

        # return (True:Bool, co:Dict[str, str])
        return (True, co)

    # construct a config object
    config = configparser.ConfigParser()

    # check if novem config file exist
    config.read(config_path)

    # if api_root is present in kwagars it overrides all other options
    co["api_root"] = ""
    if "api_root" in kwargs and kwargs["api_root"]:
        co["api_root"] = kwargs["api_root"]

    # the configuration file has an invalid format
    try:
        general = config["general"]
        profile = general["profile"]
        if "api_root" in general:
            co["api_root"] = general["api_root"]
    except KeyError:
        return (False, co)

    # override profile
    if "profile" in kwargs and kwargs["profile"]:
        profile = kwargs["profile"]

    # get our config
    try:
        uc = config[f"profile:{profile}"]
        if "api_root" in uc:
            co["api_root"] = uc["api_root"]

        co["token"] = uc["token"]
        co["username"] = uc["username"]

        if "ignore_ssl_warn" in uc:
            co["ignore_ssl_warn"] = uc.getboolean("ignore_ssl_warn")

    except KeyError:
        return (True, co)

    # kwargs supercedes
    if "api_root" in kwargs and kwargs["api_root"]:
        co["api_root"] = kwargs["api_root"]

    if "token" in kwargs and kwargs["token"]:
        co["token"] = kwargs["token"]

    return (True, co)


def pretty_format(
    values: List[Dict[str, str]], order: List[Dict[str, Any]]
) -> str:
    """
    Constructs a pretty print table of the values in values
    in the order of List
    """

    colors()

    # lets' get total terminal width (we use 120 as default)
    try:
        (col, row) = os.get_terminal_size()
    except OSError:
        col = 120

    col = col - 2

    # padding width
    pw = 2

    # unicode aware string length https://stackoverflow.com/questions/33351599/
    def ucl(word: str) -> int:
        if not word:
            return 0
        return sum(1 for ch in word if unicodedata.combining(ch) == 0)

    # construct width map
    wm = {}
    for o in order:
        k = o["key"]
        try:
            cand = max([ucl(x[k]) for x in values])
        except ValueError:
            cand = 0

        wm[k] = max([cand, len(o["header"])])

    # let's calculate our actual widths
    if sum(wm.values()) + (len(order) - 1) * pw > col:
        # we need to adjust our sizing
        ainst = {}

        nts = 0
        for o in order:
            aos = wm[o["key"]]
            if o["overflow"] == "keep" and nts + aos < col:
                ainst[o["key"]] = "keep"
                nts += aos

        rem = col - nts - ((len(order) - 1) * pw)

        # allocate truncate width
        nks = [x for x in order if x["overflow"] != "keep"]
        for o in nks:
            wm[o["key"]] = int(rem / len(nks))

    # construct output string
    los = f"{cl.BOLD}"
    for o in order:
        w = f':<{wm[o["key"]]}'
        fmt = "{0" + w + "}"
        los += fmt.format(o["header"]) + " " * pw

    los += f"{cl.ENDC}\n"
    # sep
    for o in order:
        w = f':<{wm[o["key"]]}'
        fmt = "{0" + w + "}"
        # los += fmt.format("┄" * wm[o["key"]]) + " " * pw
        los += fmt.format("╌" * wm[o["key"]]) + " " * pw

    los += "\n"

    i = 0
    for p in values:
        for o in order:
            w = f':<{wm[o["key"]]}'
            fmt = "{0" + w + "}"
            vs = wm[o["key"]]
            ov = p[o["key"]]
            if ov is None:
                ov = ""

            if len(ov) > vs:
                val = ov[0 : vs - 3] + "..."
            else:
                val = ov[0:vs]

            if "fmt" in o:
                val = o["fmt"](val)

            val = fmt.format(val)

            if "clr" in o:
                if i % 2 == 0:
                    val = f'{o["clr"]}{val}{cl.ENDC}{cl.BGGRAY}'
                else:
                    val = f'{o["clr"]}{val}{cl.ENDC}'

            if o == order[-1]:
                pad = ""
            else:
                pad = " " * pw

            if i % 2 == 0:
                los += f"{cl.BGGRAY}" + val + pad + f"{cl.ENDC}"
            else:
                los += val + pad

        los += "\n"
        i += 1

    return los


def data_on_stdin() -> Tuple[bool, str]:
    """
    identify if there is data waiting on sys.stdin
    """

    has_data = False

    try:
        # use msvcrt on windows
        import msvcrt

        if msvcrt.kbhit():  # type: ignore
            has_data = True
    except ImportError:
        try:
            # use select on linux
            has_data = bool(
                select.select(
                    [
                        sys.stdin,
                    ],
                    [],
                    [],
                    0.0,
                )[0]
            )
        except io.UnsupportedOperation:
            # We're going to assume that this is the pytest wrapper
            # if sys.stdin is an instance of io.StringIO we are mocking data
            # on stdin, so it should be true. else ignore.
            has_data = isinstance(sys.stdin, io.StringIO)

    ctnt = ""
    if has_data:
        ctnt = "".join([x for x in sys.stdin])
        if ctnt == "":
            ctnt = ""
            has_data = False

    return has_data, ctnt
